import { insertCompletionText } from '@codemirror/autocomplete';
import { syntaxTree } from '@codemirror/language';

var ELScalar;
(function (ELScalar) {
    ELScalar["Bool"] = "bool";
    ELScalar["Number"] = "number";
    ELScalar["String"] = "string";
    ELScalar["Null"] = "null";
    ELScalar["Any"] = "any";
})(ELScalar || (ELScalar = {}));

// @ts-ignore
const t = {
    deserialize: (str) => str,
};

const resolveCallable = (identifier, config) => { var _a; return (_a = config === null || config === void 0 ? void 0 : config.functions) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier); };
const resolveIdentifier = (nodeName, identifier, config) => {
    var _a;
    switch (nodeName) {
        case 'Method':
        case 'Function':
            return resolveCallable(identifier, config);
        case 'Property':
        case 'Variable':
            return (_a = config === null || config === void 0 ? void 0 : config.identifiers) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier);
    }
};
function resolveTypes(state, node, config, matchExact) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
    let types = new Set();
    if (!node) {
        return types;
    }
    let type;
    if (typeof (type = node.type.prop(t)) !== "undefined") {
        types.add(type);
    }
    else if (node.name === 'Call' && node.firstChild && node.lastChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'Variable') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_b = (_a = resolveIdentifier(node.name, varName, config)) === null || _a === void 0 ? void 0 : _a.type) === null || _b === void 0 ? void 0 : _b.forEach((x) => types.add(x));
    }
    else if (node.name === 'Function') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_d = (_c = resolveIdentifier(node.name, varName, config)) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_e = node.lastChild) === null || _e === void 0 ? void 0 : _e.name) === 'Property') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_f = resolveTypes(state, node.firstChild, config)) === null || _f === void 0 ? void 0 : _f.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.type) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_g = node.lastChild) === null || _g === void 0 ? void 0 : _g.name) === 'Method') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_h = resolveTypes(state, node.firstChild, config)) === null || _h === void 0 ? void 0 : _h.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'Application' && node.firstChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'TernaryExpression' && node.firstChild && node.firstChild.nextSibling && node.firstChild.nextSibling.nextSibling) {
        resolveTypes(state, node.firstChild.nextSibling, config).forEach(x => types.add(x));
        resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
    }
    else if (node.name === 'BinaryExpression' && ((_j = node.firstChild) === null || _j === void 0 ? void 0 : _j.nextSibling) && ((_l = (_k = node.firstChild) === null || _k === void 0 ? void 0 : _k.nextSibling) === null || _l === void 0 ? void 0 : _l.nextSibling)) {
        const operator = state.sliceDoc(node.firstChild.nextSibling.from, node.firstChild.nextSibling.to);
        if (operator == '?:' || operator == '??' || operator == '?') {
            if (operator == '?:' || operator == '??') {
                resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
            }
            resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
        }
        else if (['||', '&&', '==', '!=', '===', '!==', '>=', '<=', '>', '<'].includes(operator) || keywords.find(x => x.name == operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['**', '|', '^', '&', '<<', '>>', '+', '-', '*', '/', '%'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    else if (node.name === 'UnaryExpression' && node.firstChild) {
        const operator = state.sliceDoc(node.firstChild.from, node.firstChild.to);
        if (['not', '!'].includes(operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['+', '-'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    if (types.size === 0) {
        types.add(ELScalar.Any);
    }
    return types;
}
function getExpressionLanguageConfig(state) {
    return state.languageDataAt('expressionLanguageConfig', 0)[0];
}
const keywords = [
    { name: 'starts with', info: 'Check if a string starts with a specific string' },
    { name: 'ends with', info: 'Check if a string ends with a specific string' },
    { name: 'contains', info: 'Check if a string is not included in another string' },
    { name: 'matches', info: 'Check if a string matches a regex pattern' },
    { name: 'not in', info: 'Check if a value is not included in an array' },
    { name: 'in', info: 'Check if a value is included in an array' },
    { name: 'not' },
    { name: 'or' },
    { name: 'and' },
    { name: 'xor' },
];

const autocompleteFunction = (x) => {
    var _a, _b, _c;
    return ({
        label: `${x.name}(${((_b = (_a = x.args) === null || _a === void 0 ? void 0 : _a.map(x => x.name)) === null || _b === void 0 ? void 0 : _b.join(',')) || ''})`,
        apply: (view, completion, from, to) => {
            var _a;
            view.dispatch(Object.assign(Object.assign({}, insertCompletionText(view.state, `${x.name}()`, from, to)), { selection: { anchor: from + x.name.length + (((_a = x.args) === null || _a === void 0 ? void 0 : _a.length) > 0 ? 1 : 2) } }));
        },
        detail: (_c = x.returnType) === null || _c === void 0 ? void 0 : _c.join('|'),
        info: x.info,
        type: "function",
    });
};
const autocompleteIdentifier = (x) => {
    var _a;
    return ({
        label: x.name,
        apply: x.name,
        info: x.info,
        detail: x.detail || ((_a = x.type) === null || _a === void 0 ? void 0 : _a.join('|')),
        type: 'variable',
    });
};
function completeOperatorKeyword(state, config, tree, from, to, explicit) {
    var _a;
    return {
        from,
        to,
        options: (_a = keywords.map(({ name, info, detail }) => ({
            label: name,
            apply: `${name} `,
            info: info,
            detail,
            type: "keyword"
        }))) !== null && _a !== void 0 ? _a : [],
        validFor: (text) => { var _a; return (_a = keywords.some(({ name }) => explicit || name.includes(text))) !== null && _a !== void 0 ? _a : false; },
    };
}
function completeIdentifier(state, config, tree, from, to) {
    var _a, _b;
    const identifiers = (_a = config.identifiers) !== null && _a !== void 0 ? _a : []; //?.filter(({ name }) => explicit || name.toLowerCase().startsWith(text)) ?? [];
    const functions = (_b = config.functions) !== null && _b !== void 0 ? _b : []; //?.filter(({ name }) => explicit || name.toLowerCase().startsWith(text)) ?? [];
    return {
        from,
        to,
        options: [...(identifiers.map(autocompleteIdentifier)), ...(functions.map(autocompleteFunction))],
        validFor: /^[a-zA-Z_]+[a-zA-Z_0-9]*$/,
    };
}
function completeMember(state, config, tree, from, to, explicit) {
    var _a, _b, _c, _d;
    if (((_a = tree.parent) === null || _a === void 0 ? void 0 : _a.name) != 'ObjectAccess' || !tree.parent.firstChild) {
        return null;
    }
    const types = resolveTypes(state, tree.parent.firstChild.node, config);
    if (!(types === null || types === void 0 ? void 0 : types.size)) {
        return null;
    }
    let options = [];
    for (const type of types) {
        const typeDeclaration = (_b = config.types) === null || _b === void 0 ? void 0 : _b[type];
        options.push(...(((_c = typeDeclaration === null || typeDeclaration === void 0 ? void 0 : typeDeclaration.identifiers) === null || _c === void 0 ? void 0 : _c.map(autocompleteIdentifier)) || []), ...(((_d = typeDeclaration === null || typeDeclaration === void 0 ? void 0 : typeDeclaration.functions) === null || _d === void 0 ? void 0 : _d.map(autocompleteFunction)) || []));
    }
    return {
        from,
        to,
        options,
        validFor: /^[a-zA-Z_]+[a-zA-Z_0-9]*$/,
    };
}
function expressionLanguageCompletion(context) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o;
    const { state, pos, explicit } = context;
    const tree = syntaxTree(state);
    const lastChar = state.sliceDoc(pos - 1, pos);
    const prevNode = tree.resolveInner(pos, lastChar === ')' ? 0 : -1);
    const config = getExpressionLanguageConfig(state);
    const isIdentifier = (node) => { var _a; return ['Variable', 'Function'].includes((_a = node === null || node === void 0 ? void 0 : node.name) !== null && _a !== void 0 ? _a : ''); };
    const isMember = (node) => { var _a; return ['Property', 'Method'].includes((_a = node === null || node === void 0 ? void 0 : node.name) !== null && _a !== void 0 ? _a : ''); };
    if (prevNode.name == 'String' || prevNode.name == 'BlockComment') {
        return null;
    }
    if (((_a = prevNode.parent) === null || _a === void 0 ? void 0 : _a.name) == 'ObjectAccess' && ['ObjectAccess', 'ArrayAccess', 'Variable', 'Call', 'Application'].includes(((_b = prevNode.parent.firstChild) === null || _b === void 0 ? void 0 : _b.name) || '')) {
        return completeMember(state, config, prevNode, isIdentifier(prevNode) || isMember(prevNode) ? prevNode.from : pos, pos);
    }
    if (['Expression', 'UnaryExpression', 'BinaryExpression', 'TernaryExpression'].includes(prevNode.name) && prevNode.lastChild && !((_c = prevNode.lastChild) === null || _c === void 0 ? void 0 : _c.type.isError)
        || ['Arguments', 'Array'].includes(prevNode.name) && prevNode.lastChild && !((_d = prevNode.lastChild) === null || _d === void 0 ? void 0 : _d.type.isError)
        || ['Expression', 'UnaryExpression', 'BinaryExpression', 'TernaryExpression'].includes((_f = (_e = prevNode.parent) === null || _e === void 0 ? void 0 : _e.name) !== null && _f !== void 0 ? _f : '') && prevNode.type.isError
        || ['Variable', 'Function'].includes((_h = (_g = prevNode.parent) === null || _g === void 0 ? void 0 : _g.name) !== null && _h !== void 0 ? _h : '') && prevNode.type.isError) {
        return completeOperatorKeyword(state, config, prevNode, !['Expression', 'UnaryExpression', 'BinaryExpression', 'TernaryExpression', 'Arguments'].includes(prevNode.name) ? prevNode.from : pos, pos, explicit);
    }
    if (!/[0-9]/.test(lastChar) && !['OperatorKeyword'].includes((_j = prevNode.name) !== null && _j !== void 0 ? _j : '') && (['Expression', 'UnaryExpression', 'BinaryExpression', 'TernaryExpression'].includes(prevNode.name) && ((_k = prevNode.lastChild) === null || _k === void 0 ? void 0 : _k.type.isError)
        || ['Expression', 'UnaryExpression', 'BinaryExpression', 'TernaryExpression', 'Arguments'].includes((_m = (_l = prevNode.parent) === null || _l === void 0 ? void 0 : _l.name) !== null && _m !== void 0 ? _m : '') && !prevNode.type.isError
        || ['Arguments', 'Array'].includes((_o = prevNode.name) !== null && _o !== void 0 ? _o : ''))) {
        return completeIdentifier(state, config, prevNode, isIdentifier(prevNode) ? prevNode.from : pos, pos);
    }
    return null;
}

export { expressionLanguageCompletion };
