import { linter } from '@codemirror/lint';
import { syntaxTree } from '@codemirror/language';

var ELScalar;
(function (ELScalar) {
    ELScalar["Bool"] = "bool";
    ELScalar["Number"] = "number";
    ELScalar["String"] = "string";
    ELScalar["Null"] = "null";
    ELScalar["Any"] = "any";
})(ELScalar || (ELScalar = {}));

// @ts-ignore
const t = {
    deserialize: (str) => str,
};

function resolveFunctionDefinition(node, state, config) {
    var _a;
    if (!node) {
        return undefined;
    }
    let identifier;
    if (node.name === 'ObjectAccess' && node.lastChild) {
        const leftArgument = (_a = node.firstChild) === null || _a === void 0 ? void 0 : _a.node;
        const types = Array.from(resolveTypes(state, leftArgument, config));
        identifier = state.sliceDoc(node.lastChild.from, node.lastChild.to);
        return types.map(type => { var _a; return resolveCallable(identifier, (_a = config.types) === null || _a === void 0 ? void 0 : _a[type]); }).find(x => x);
    }
    else if (node.name === 'Function') {
        identifier = state.sliceDoc(node.from, node.node.firstChild ? node.node.firstChild.from - 1 : node.to);
        return resolveCallable(identifier, config);
    }
}
const resolveCallable = (identifier, config) => { var _a; return (_a = config === null || config === void 0 ? void 0 : config.functions) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier); };
const resolveIdentifier = (nodeName, identifier, config) => {
    var _a;
    switch (nodeName) {
        case 'Method':
        case 'Function':
            return resolveCallable(identifier, config);
        case 'Property':
        case 'Variable':
            return (_a = config === null || config === void 0 ? void 0 : config.identifiers) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier);
    }
};
function resolveTypes(state, node, config, matchExact) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
    let types = new Set();
    if (!node) {
        return types;
    }
    let type;
    if (typeof (type = node.type.prop(t)) !== "undefined") {
        types.add(type);
    }
    else if (node.name === 'Call' && node.firstChild && node.lastChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'Variable') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_b = (_a = resolveIdentifier(node.name, varName, config)) === null || _a === void 0 ? void 0 : _a.type) === null || _b === void 0 ? void 0 : _b.forEach((x) => types.add(x));
    }
    else if (node.name === 'Function') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_d = (_c = resolveIdentifier(node.name, varName, config)) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_e = node.lastChild) === null || _e === void 0 ? void 0 : _e.name) === 'Property') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_f = resolveTypes(state, node.firstChild, config)) === null || _f === void 0 ? void 0 : _f.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.type) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_g = node.lastChild) === null || _g === void 0 ? void 0 : _g.name) === 'Method') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_h = resolveTypes(state, node.firstChild, config)) === null || _h === void 0 ? void 0 : _h.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'Application' && node.firstChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'TernaryExpression' && node.firstChild && node.firstChild.nextSibling && node.firstChild.nextSibling.nextSibling) {
        resolveTypes(state, node.firstChild.nextSibling, config).forEach(x => types.add(x));
        resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
    }
    else if (node.name === 'BinaryExpression' && ((_j = node.firstChild) === null || _j === void 0 ? void 0 : _j.nextSibling) && ((_l = (_k = node.firstChild) === null || _k === void 0 ? void 0 : _k.nextSibling) === null || _l === void 0 ? void 0 : _l.nextSibling)) {
        const operator = state.sliceDoc(node.firstChild.nextSibling.from, node.firstChild.nextSibling.to);
        if (operator == '?:' || operator == '??' || operator == '?') {
            if (operator == '?:' || operator == '??') {
                resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
            }
            resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
        }
        else if (['||', '&&', '==', '!=', '===', '!==', '>=', '<=', '>', '<'].includes(operator) || keywords.find(x => x.name == operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['**', '|', '^', '&', '<<', '>>', '+', '-', '*', '/', '%'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    else if (node.name === 'UnaryExpression' && node.firstChild) {
        const operator = state.sliceDoc(node.firstChild.from, node.firstChild.to);
        if (['not', '!'].includes(operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['+', '-'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    if (types.size === 0) {
        types.add(ELScalar.Any);
    }
    return types;
}
function getExpressionLanguageConfig(state) {
    return state.languageDataAt('expressionLanguageConfig', 0)[0];
}
const keywords = [
    { name: 'starts with', info: 'Check if a string starts with a specific string' },
    { name: 'ends with', info: 'Check if a string ends with a specific string' },
    { name: 'contains', info: 'Check if a string is not included in another string' },
    { name: 'matches', info: 'Check if a string matches a regex pattern' },
    { name: 'not in', info: 'Check if a value is not included in an array' },
    { name: 'in', info: 'Check if a value is included in an array' },
    { name: 'not' },
    { name: 'or' },
    { name: 'and' },
    { name: 'xor' },
];

/**
 * @internal
 */
const expressionLanguageLinterSource = (state) => {
    const config = getExpressionLanguageConfig(state);
    let diagnostics = [];
    syntaxTree(state).cursor().iterate(node => {
        var _a, _b, _c, _d, _e, _f, _g;
        const { from, to, name } = node;
        let identifier;
        switch (name) {
            case '⚠':
                if (state.doc.length === 0 || from === 0) {
                    // Don't show error on empty doc (even though it is an error)
                    return;
                }
                identifier = state.sliceDoc(from, to);
                if (identifier.length === 0) {
                    diagnostics.push({ from, to: (_c = (_b = (_a = node.node.parent) === null || _a === void 0 ? void 0 : _a.parent) === null || _b === void 0 ? void 0 : _b.to) !== null && _c !== void 0 ? _c : to, severity: 'error', message: `Expression expected` });
                }
                else {
                    const type = /^[a-zA-Z_]+[a-zA-Z_0-9]*$/.test(identifier) ? 'identifier' : 'operator';
                    diagnostics.push({ from, to, severity: 'error', message: `Unexpected ${type} '${identifier}'` });
                }
                return;
            case 'Arguments':
                const args = (_d = resolveFunctionDefinition(node.node.prevSibling, state, config)) === null || _d === void 0 ? void 0 : _d.args;
                if (!args) {
                    return;
                }
                for (let n = node.node.firstChild, i = 0; n != null; n = n.nextSibling) {
                    if (n.name === 'BlockComment') {
                        continue;
                    }
                    if (i > args.length - 1) {
                        diagnostics.push({ from: n.from, to: n.to, severity: 'error', message: `Unexpected argument` });
                        continue;
                    }
                    const typesUsed = Array.from(resolveTypes(state, n, config));
                    const typesExpected = args[i].type;
                    if (typesExpected && !typesExpected.includes(ELScalar.Any) && !typesUsed.some(x => typesExpected.includes(x))) {
                        diagnostics.push({ from: n.from, to: n.to, severity: 'error', message: `<code>${typesExpected.join('|')}</code> expected, got <code>${typesUsed.join('|')}</code>` });
                    }
                    i++;
                }
                break;
            case 'Property':
            case 'Method':
                const leftArgument = (_f = (_e = node.node.parent) === null || _e === void 0 ? void 0 : _e.firstChild) === null || _f === void 0 ? void 0 : _f.node;
                const types = Array.from(resolveTypes(state, leftArgument, config));
                identifier = state.sliceDoc(from, to);
                if (!types.find(type => { var _a; return resolveIdentifier(name, identifier, (_a = config.types) === null || _a === void 0 ? void 0 : _a[type]); })) {
                    diagnostics.push({ from, to, severity: 'error', message: `${node.name} "${identifier}" not found in ${types.join('|')}` });
                }
                break;
            case 'Variable':
            case 'Function':
                identifier = state.sliceDoc(from, node.node.firstChild ? node.node.firstChild.from - 1 : to);
                if (!resolveIdentifier(name, identifier, config)) {
                    diagnostics.push({ from, to, severity: 'error', message: `${node.node.name} "${identifier}" not found` });
                }
                break;
        }
        if (identifier && ((_g = node.node.parent) === null || _g === void 0 ? void 0 : _g.type.isError)) {
            diagnostics.push({ from, to, severity: 'error', message: `Unexpected identifier "${identifier}"` });
        }
    });
    diagnostics.forEach(d => {
        d.renderMessage = () => {
            const span = document.createElement('span');
            span.innerHTML = d.message;
            return span;
        };
    });
    return diagnostics;
};
const expressionLanguageLinter = linter(view => expressionLanguageLinterSource(view.state));

export { expressionLanguageLinter, expressionLanguageLinterSource };
