'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var state = require('@codemirror/state');
var view = require('@codemirror/view');
var language = require('@codemirror/language');

var ELScalar;
(function (ELScalar) {
    ELScalar["Bool"] = "bool";
    ELScalar["Number"] = "number";
    ELScalar["String"] = "string";
    ELScalar["Null"] = "null";
    ELScalar["Any"] = "any";
})(ELScalar || (ELScalar = {}));

// @ts-ignore
const t = {
    deserialize: (str) => str,
};

const createInfoElement = (html) => {
    const dom = document.createElement("div");
    dom.innerHTML = html;
    dom.className = 'cm-diagnostic';
    return dom;
};
function resolveFunctionDefinition(node, state, config) {
    var _a;
    if (!node) {
        return undefined;
    }
    let identifier;
    if (node.name === 'ObjectAccess' && node.lastChild) {
        const leftArgument = (_a = node.firstChild) === null || _a === void 0 ? void 0 : _a.node;
        const types = Array.from(resolveTypes(state, leftArgument, config));
        identifier = state.sliceDoc(node.lastChild.from, node.lastChild.to);
        return types.map(type => { var _a; return resolveCallable(identifier, (_a = config.types) === null || _a === void 0 ? void 0 : _a[type]); }).find(x => x);
    }
    else if (node.name === 'Function') {
        identifier = state.sliceDoc(node.from, node.node.firstChild ? node.node.firstChild.from - 1 : node.to);
        return resolveCallable(identifier, config);
    }
}
const resolveCallable = (identifier, config) => { var _a; return (_a = config === null || config === void 0 ? void 0 : config.functions) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier); };
const resolveIdentifier = (nodeName, identifier, config) => {
    var _a;
    switch (nodeName) {
        case 'Method':
        case 'Function':
            return resolveCallable(identifier, config);
        case 'Property':
        case 'Variable':
            return (_a = config === null || config === void 0 ? void 0 : config.identifiers) === null || _a === void 0 ? void 0 : _a.find(x => x.name === identifier);
    }
};
function resolveTypes(state, node, config, matchExact) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l;
    let types = new Set();
    if (!node) {
        return types;
    }
    let type;
    if (typeof (type = node.type.prop(t)) !== "undefined") {
        types.add(type);
    }
    else if (node.name === 'Call' && node.firstChild && node.lastChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'Variable') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_b = (_a = resolveIdentifier(node.name, varName, config)) === null || _a === void 0 ? void 0 : _a.type) === null || _b === void 0 ? void 0 : _b.forEach((x) => types.add(x));
    }
    else if (node.name === 'Function') {
        const varName = state.sliceDoc(node.from, node.to) || '';
        // @ts-ignore
        (_d = (_c = resolveIdentifier(node.name, varName, config)) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_e = node.lastChild) === null || _e === void 0 ? void 0 : _e.name) === 'Property') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_f = resolveTypes(state, node.firstChild, config)) === null || _f === void 0 ? void 0 : _f.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.type) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'ObjectAccess' && node.firstChild && ((_g = node.lastChild) === null || _g === void 0 ? void 0 : _g.name) === 'Method') {
        const varName = state.sliceDoc(node.lastChild.from, node.lastChild.to) || '';
        (_h = resolveTypes(state, node.firstChild, config)) === null || _h === void 0 ? void 0 : _h.forEach(baseType => {
            var _a, _b, _c, _d;
            // @ts-ignore
            (_d = (_c = resolveIdentifier((_a = node.lastChild) === null || _a === void 0 ? void 0 : _a.name, varName, (_b = config.types) === null || _b === void 0 ? void 0 : _b[baseType])) === null || _c === void 0 ? void 0 : _c.returnType) === null || _d === void 0 ? void 0 : _d.forEach((x) => types.add(x));
        });
    }
    else if (node.name === 'Application' && node.firstChild) {
        resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
    }
    else if (node.name === 'TernaryExpression' && node.firstChild && node.firstChild.nextSibling && node.firstChild.nextSibling.nextSibling) {
        resolveTypes(state, node.firstChild.nextSibling, config).forEach(x => types.add(x));
        resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
    }
    else if (node.name === 'BinaryExpression' && ((_j = node.firstChild) === null || _j === void 0 ? void 0 : _j.nextSibling) && ((_l = (_k = node.firstChild) === null || _k === void 0 ? void 0 : _k.nextSibling) === null || _l === void 0 ? void 0 : _l.nextSibling)) {
        const operator = state.sliceDoc(node.firstChild.nextSibling.from, node.firstChild.nextSibling.to);
        if (operator == '?:' || operator == '??' || operator == '?') {
            if (operator == '?:' || operator == '??') {
                resolveTypes(state, node.firstChild, config).forEach(x => types.add(x));
            }
            resolveTypes(state, node.firstChild.nextSibling.nextSibling, config).forEach(x => types.add(x));
        }
        else if (['||', '&&', '==', '!=', '===', '!==', '>=', '<=', '>', '<'].includes(operator) || keywords.find(x => x.name == operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['**', '|', '^', '&', '<<', '>>', '+', '-', '*', '/', '%'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    else if (node.name === 'UnaryExpression' && node.firstChild) {
        const operator = state.sliceDoc(node.firstChild.from, node.firstChild.to);
        if (['not', '!'].includes(operator)) {
            types.add(ELScalar.Bool);
        }
        else if (['+', '-'].includes(operator)) {
            types.add(ELScalar.Number);
        }
    }
    if (types.size === 0) {
        types.add(ELScalar.Any);
    }
    return types;
}
function getExpressionLanguageConfig(state) {
    return state.languageDataAt('expressionLanguageConfig', 0)[0];
}
const keywords = [
    { name: 'starts with', info: 'Check if a string starts with a specific string' },
    { name: 'ends with', info: 'Check if a string ends with a specific string' },
    { name: 'contains', info: 'Check if a string is not included in another string' },
    { name: 'matches', info: 'Check if a string matches a regex pattern' },
    { name: 'not in', info: 'Check if a value is not included in an array' },
    { name: 'in', info: 'Check if a value is included in an array' },
    { name: 'not' },
    { name: 'or' },
    { name: 'and' },
    { name: 'xor' },
];

function getNodeOrdinal(node) {
    let ordinal = -1;
    // noinspection StatementWithEmptyBodyJS
    for (let c = node; c; c = c.prevSibling, ordinal++)
        ;
    return ordinal;
}
function resolveArguments(node) {
    let c = node;
    while (c && c.name !== 'Arguments') {
        c = c.parent;
    }
    return c;
}
function getCursorTooltips(state) {
    const config = getExpressionLanguageConfig(state);
    // @ts-ignore
    return state.selection.ranges
        .filter(range => range.empty)
        .map(range => {
        var _a, _b;
        const tree = language.syntaxTree(state);
        const node = tree.resolveInner(range.from, 0);
        const args = resolveArguments(node);
        if (!args || !args.prevSibling) {
            return null;
        }
        const fn = resolveFunctionDefinition(args.prevSibling, state, config);
        if (!fn) {
            return null;
        }
        const n = args.childAfter(range.from - 1);
        const argName = (_b = (_a = fn.args) === null || _a === void 0 ? void 0 : _a[n ? getNodeOrdinal(n) : 0]) === null || _b === void 0 ? void 0 : _b.name;
        if (n && n.from !== range.from || !argName) {
            return null;
        }
        return {
            pos: range.head,
            above: true,
            strictSide: false,
            arrow: true,
            create: () => {
                let dom = document.createElement("div");
                dom.className = "cm-tooltip-cursor";
                dom.textContent = `${argName}`;
                return { dom };
            },
        };
    }).filter(x => x);
}
const cursorTooltipField = state.StateField.define({
    create: (state) => getCursorTooltips(state),
    update(tooltips, tr) {
        if (!tr.docChanged && !tr.selection) {
            return tooltips;
        }
        return getCursorTooltips(tr.state);
    },
    provide: f => view.showTooltip.computeN([f], state => state.field(f))
});
const keywordTooltip = view.hoverTooltip((view, pos, side) => {
    var _a, _b, _c, _d, _e, _f, _g;
    const config = getExpressionLanguageConfig(view.state);
    const tree = language.syntaxTree(view.state).resolveInner(pos, side);
    if (tree.name === 'OperatorKeyword') {
        const name = view.state.sliceDoc(tree.from, tree.to);
        const info = (_a = keywords.find(x => x.name === name)) === null || _a === void 0 ? void 0 : _a.info;
        if (info) {
            return {
                pos: tree.from,
                end: tree.to,
                above: true,
                create: () => ({ dom: createInfoElement(info) }),
            };
        }
    }
    if (!['Function', 'Variable', 'Method', 'Property'].includes(tree.name)) {
        return null;
    }
    const skipEmpty = (x) => x;
    let info;
    if (((_b = tree.parent) === null || _b === void 0 ? void 0 : _b.firstChild) && ((_c = tree.parent) === null || _c === void 0 ? void 0 : _c.name) === 'ObjectAccess' && tree.prevSibling) {
        const node = tree.parent.firstChild;
        const types = Array.from(resolveTypes(view.state, node, config));
        const name = view.state.sliceDoc(tree.from, tree.to);
        info = [
            ...types.map(type => { var _a, _b, _c, _d; return (_d = (_c = (_b = (_a = config.types) === null || _a === void 0 ? void 0 : _a[type]) === null || _b === void 0 ? void 0 : _b.identifiers) === null || _c === void 0 ? void 0 : _c.find(x => x.name === name)) === null || _d === void 0 ? void 0 : _d.info; }).filter(skipEmpty),
            ...types.map(type => { var _a, _b, _c, _d; return (_d = (_c = (_b = (_a = config.types) === null || _a === void 0 ? void 0 : _a[type]) === null || _b === void 0 ? void 0 : _b.functions) === null || _c === void 0 ? void 0 : _c.find(x => x.name === name)) === null || _d === void 0 ? void 0 : _d.info; }).filter(skipEmpty),
        ].join('\n');
    }
    else {
        const name = view.state.sliceDoc(tree.from, tree.to);
        info = [
            ...[(_e = (_d = config.identifiers) === null || _d === void 0 ? void 0 : _d.find(x => x.name === name)) === null || _e === void 0 ? void 0 : _e.info].filter(skipEmpty),
            ...[(_g = (_f = config.functions) === null || _f === void 0 ? void 0 : _f.find(x => x.name === name)) === null || _g === void 0 ? void 0 : _g.info].filter(skipEmpty),
        ].join('\n');
    }
    if (!info) {
        return null;
    }
    return {
        pos: tree.from,
        end: tree.to,
        above: true,
        create: () => ({ dom: createInfoElement(info) }),
    };
});

exports.cursorTooltipField = cursorTooltipField;
exports.keywordTooltip = keywordTooltip;
