import { SyntaxNode } from "@lezer/common";
import { EditorState } from "@codemirror/state";
/**
 * Represents a variable or a property of an object
 */
interface ELIdentifier {
    name: string;
    detail?: string;
    info?: string;
    type?: string[];
}
/**
 * Represents a function or a method of an object
 */
interface ELFunction {
    name: string;
    args: ELParameter[];
    info?: string;
    returnType?: string[];
}
interface ELParameter {
    name: string;
    type?: string[];
    info?: string;
    optional?: boolean;
}
interface ELType {
    identifiers?: ELIdentifier[];
    functions?: ELFunction[];
    info?: string;
}
interface ExpressionLanguageConfig {
    types?: {
        [key: string]: ELType;
    };
    identifiers?: ELIdentifier[];
    functions?: ELFunction[];
}
interface ELKeyword {
    name: string;
    detail?: string;
    info?: string;
}
declare const createInfoElement: (html: string) => HTMLDivElement;
declare function resolveFunctionDefinition(node: SyntaxNode | null, state: EditorState, config: ExpressionLanguageConfig): ELFunction | undefined;
declare const resolveIdentifier: (nodeName: 'Method' | 'Property' | 'Function' | 'Variable', identifier?: string, config?: {
    identifiers?: ELIdentifier[];
    functions?: ELFunction[];
}) => ELIdentifier | ELFunction | undefined;
declare function resolveTypes(state: EditorState, node: SyntaxNode | undefined, config: ExpressionLanguageConfig, matchExact: boolean): Set<string>;
declare function getExpressionLanguageConfig(state: EditorState): ExpressionLanguageConfig;
declare const keywords: ELKeyword[];
export { createInfoElement, resolveFunctionDefinition, resolveIdentifier, resolveTypes, getExpressionLanguageConfig, keywords };
